@@ -433,6 +433,10 @@ gdk_pixbuf_io_init (void)
 				have_error = TRUE;
 			}			
 			module->info->description = g_strdup (tmp_buf->str);
+
+			if (scan_string (&p, tmp_buf)) {
+				module->info->license = g_strdup (tmp_buf->str);
+			}
 		}
 		else if (!module->info->mime_types) {
 			int n = 1;
@@ -975,6 +979,9 @@ size_prepared_cb (GdkPixbufLoader *loader,
 			height = info->height;
 	}
 	
+	width = MAX (width, 1);
+        height = MAX (height, 1);
+
 	gdk_pixbuf_loader_set_size (loader, width, height);
 }
 
@@ -1297,6 +1304,8 @@ gdk_pixbuf_new_from_xpm_data (const char **data)
 	GdkPixbufModule *xpm_module;
 	gboolean locked;
 
+	g_return_val_if_fail (data != NULL, NULL);
+
 	xpm_module = _gdk_pixbuf_get_named_module ("xpm", &error);
 	if (xpm_module == NULL) {
 		g_warning ("Error loading XPM image loader: %s", error->message);
