@@ -60,6 +60,7 @@ _gdk_win32_selection_init (void)
 {
   sel_prop_table = g_hash_table_new (NULL, NULL);
   sel_owner_table = g_hash_table_new (NULL, NULL);
+  _format_atom_table = g_hash_table_new (NULL, NULL);
 }
 
 /* The specifications for COMPOUND_TEXT and STRING specify that C0 and
@@ -200,17 +201,22 @@ gdk_selection_owner_set_for_display (GdkDisplay *display,
 {
   HWND hwnd;
   GdkEvent tmp_event;
-  gchar *sel_name;
 
   g_return_val_if_fail (display == _gdk_display, FALSE);
   g_return_val_if_fail (selection != GDK_NONE, FALSE);
 
-  GDK_NOTE (DND,
-	    (sel_name = gdk_atom_name (selection),
-	     g_print ("gdk_selection_owner_set: %p %#x (%s)\n",
-		      (owner ? GDK_WINDOW_HWND (owner) : NULL),
-		      (guint) selection, sel_name),
-	     g_free (sel_name)));
+#ifdef G_ENABLE_DEBUG
+  {
+    gchar *sel_name;
+
+    GDK_NOTE (DND,
+	      (sel_name = gdk_atom_name (selection),
+	       g_print ("gdk_selection_owner_set_for_display: %p %#x (%s)\n",
+			(owner ? GDK_WINDOW_HWND (owner) : NULL),
+			(guint) selection, sel_name),
+	       g_free (sel_name)));
+  }
+#endif
 
   if (selection != GDK_SELECTION_CLIPBOARD)
     {
@@ -236,6 +242,7 @@ gdk_selection_owner_set_for_display (GdkDisplay *display,
     return FALSE;
 
   _ignore_destroy_clipboard = TRUE;
+  GDK_NOTE (DND, g_print ("... EmptyClipboard()\n"));
   if (!API_CALL (EmptyClipboard, ()))
     {
       _ignore_destroy_clipboard = FALSE;
@@ -253,7 +260,7 @@ gdk_selection_owner_set_for_display (GdkDisplay *display,
        * gdk_property_change will be called to store the clipboard
        * data.
        */
-      GDK_NOTE (DND, g_print ("...sending GDK_SELECTION_REQUEST to ourselves\n"));
+      GDK_NOTE (DND, g_print ("... sending GDK_SELECTION_REQUEST to ourselves\n"));
       tmp_event.selection.type = GDK_SELECTION_REQUEST;
       tmp_event.selection.window = owner;
       tmp_event.selection.send_event = FALSE;
@@ -274,28 +281,34 @@ gdk_selection_owner_get_for_display (GdkDisplay *display,
                                      GdkAtom     selection)
 {
   GdkWindow *window;
-  gchar *sel_name;
 
   g_return_val_if_fail (display == _gdk_display, NULL);
   g_return_val_if_fail (selection != GDK_NONE, NULL);
 
-  /* Return NULL for CLIPBOARD, because otherwise cut&paste inside the
-   * same application doesn't work. We must pretend to gtk that we
-   * don't have the selection, so that we always fetch it from the
-   * Windows clipboard. See also comments in
-   * gdk_selection_send_notify().
-   */
   if (selection == GDK_SELECTION_CLIPBOARD)
-    return NULL;
+    {
+      HWND owner = GetClipboardOwner ();
+
+      if (owner == NULL)
+	return NULL;
+
+      return gdk_win32_handle_table_lookup ((GdkNativeWindow) owner);
+    }
 
   window = gdk_window_lookup ((GdkNativeWindow) g_hash_table_lookup (sel_owner_table, selection));
 
-  GDK_NOTE (DND,
-	    (sel_name = gdk_atom_name (selection),
-	     g_print ("gdk_selection_owner_get: %#x (%s) = %p\n",
-		      (guint) selection, sel_name,
-		      (window ? GDK_WINDOW_HWND (window) : NULL)),
-	     g_free (sel_name)));
+#ifdef G_ENABLE_DEBUG
+  {
+    gchar *sel_name;
+    
+    GDK_NOTE (DND,
+	      (sel_name = gdk_atom_name (selection),
+	       g_print ("gdk_selection_owner_get: %#x (%s) = %p\n",
+			(guint) selection, sel_name,
+			(window ? GDK_WINDOW_HWND (window) : NULL)),
+	       g_free (sel_name)));
+  }
+#endif
 
   return window;
 }
@@ -329,8 +342,6 @@ gdk_selection_convert (GdkWindow *requestor,
 {
   HGLOBAL hdata;
   GdkAtom property = _gdk_selection_property;
-  gchar *sel_name, *tgt_name;
-  GError *error = NULL;
 
   g_return_if_fail (selection != GDK_NONE);
   g_return_if_fail (requestor != NULL);
@@ -338,72 +349,81 @@ gdk_selection_convert (GdkWindow *requestor,
   if (GDK_WINDOW_DESTROYED (requestor))
     return;
 
-  GDK_NOTE (DND,
-	    (sel_name = gdk_atom_name (selection),
-	     tgt_name = gdk_atom_name (target),
-	     g_print ("gdk_selection_convert: %p %#x (%s) %#x (%s)\n",
-		      GDK_WINDOW_HWND (requestor),
-		      (guint) selection, sel_name,
-		      (guint) target, tgt_name),
-	     g_free (sel_name),
-	     g_free (tgt_name)));
+#ifdef G_ENABLE_DEBUG
+  {
+    gchar *sel_name, *tgt_name;
+    
+    GDK_NOTE (DND,
+	      (sel_name = gdk_atom_name (selection),
+	       tgt_name = gdk_atom_name (target),
+	       g_print ("gdk_selection_convert: %p %#x (%s) %#x (%s)\n",
+			GDK_WINDOW_HWND (requestor),
+			(guint) selection, sel_name,
+			(guint) target, tgt_name),
+	       g_free (sel_name),
+	       g_free (tgt_name)));
+  }
+#endif
 
   if (selection == GDK_SELECTION_CLIPBOARD && target == _targets)
     {
-      /* He wants to know what formats are on the clipboard.  If there
+      gint formats_cnt, i, fmt;
+      GdkAtom *data;
+      gboolean has_bmp = FALSE;
+
+      /* He wants to know what formats are on the clipboard. If there
        * is some kind of text, tell him so.
        */
       if (!API_CALL (OpenClipboard, (GDK_WINDOW_HWND (requestor))))
 	return;
 
+      formats_cnt = CountClipboardFormats ();
+      data = g_new (GdkAtom, formats_cnt + 2);
+      i = 0;
+
       if (IsClipboardFormatAvailable (CF_UNICODETEXT) ||
 	  IsClipboardFormatAvailable (_cf_utf8_string) ||
 	  IsClipboardFormatAvailable (CF_TEXT))
 	{
-	  GdkAtom *data = g_new (GdkAtom, 1);
-	  *data = _utf8_string;
-	  _gdk_selection_property_store (requestor, GDK_SELECTION_TYPE_ATOM,
-					 32, (guchar *) data, 1 * sizeof (GdkAtom));
+	  data[i++] = _utf8_string;
 	}
-      else if (IsClipboardFormatAvailable (CF_BITMAP) ||
-               IsClipboardFormatAvailable (CF_DIB))
-	{
-	  GdkAtom *data = g_new (GdkAtom, 1);
-	  GdkAtom atom = gdk_atom_intern ("image/bmp", FALSE);
-          *data = atom;
-	  _gdk_selection_property_store (requestor, GDK_SELECTION_TYPE_ATOM,
-					 32, (guchar *) data, 1 * sizeof (GdkAtom));
-	}
-      else if (CountClipboardFormats() > 0)
+      if (formats_cnt > 0)
         {
-          /* if there is anything else in the clipboard, enum it all although we don't 
-           * offer special conversion services 
+          /* If there is anything else in the clipboard, enum it all
+           * although we don't offer special conversion services.
            */
-          int fmt = 0, i = 0;
-	  GdkAtom *data = g_new (GdkAtom, CountClipboardFormats());
-
-          for ( ; 0 != (fmt = EnumClipboardFormats (fmt)); )
+          for (fmt = 0; 0 != (fmt = EnumClipboardFormats (fmt)); )
             {
-              char sFormat[80];
+              gchar sFormat[80];
 
-              if (GetClipboardFormatName (fmt, sFormat, 80) > 0)
+              if (GetClipboardFormatName (fmt, sFormat, 80) > 0 &&
+		  strcmp (sFormat, "UTF8_STRING"))
                 {
-                  GdkAtom atom = gdk_atom_intern (sFormat, FALSE);
-                  data[i] = atom;
-                  i++;
+		  GdkAtom atom;
+
+		  if (!has_bmp &&
+		      (!strcmp (sFormat, "image/bmp") ||
+		       !strcmp (sFormat, "image/x-bmp") ||
+		       !strcmp (sFormat, "image/x-MS-bmp")))
+		    has_bmp = TRUE;
+                  atom = gdk_atom_intern (sFormat, FALSE);
+                  data[i++] = atom;
                 }
             }
-	  _gdk_selection_property_store (requestor, GDK_SELECTION_TYPE_ATOM,
-					 32, (guchar *) data, i * sizeof (GdkAtom));
         }
+      if (!has_bmp && (IsClipboardFormatAvailable (CF_BITMAP) ||
+		       IsClipboardFormatAvailable (CF_DIB)))
+	data[i++] = _image_bmp;
+
+      if (i > 0)
+	_gdk_selection_property_store (requestor, GDK_SELECTION_TYPE_ATOM,
+				       32, (guchar *) data, i * sizeof (GdkAtom));
       else             
 	property = GDK_NONE;
 
       API_CALL (CloseClipboard, ());
     }
-  else if (selection == GDK_SELECTION_CLIPBOARD &&
-	   (target == GDK_TARGET_STRING ||
-	    target == _utf8_string))
+  else if (selection == GDK_SELECTION_CLIPBOARD && target == _utf8_string)
     {
       /* Converting the CLIPBOARD selection means he wants the
        * contents of the clipboard. Get the clipboard data, and store
@@ -413,7 +433,7 @@ gdk_selection_convert (GdkWindow *requestor,
 	return;
 
       /* Try various formats. First the simplest, CF_UNICODETEXT. */
-      if (G_WIN32_IS_NT_BASED () && (hdata = GetClipboardData (CF_UNICODETEXT)) != NULL)
+      if ((hdata = GetClipboardData (CF_UNICODETEXT)) != NULL)
 	{
 	  wchar_t *ptr, *wcs, *p, *q;
 	  guchar *data;
@@ -423,7 +443,7 @@ gdk_selection_convert (GdkWindow *requestor,
 	    {
 	      length = GlobalSize (hdata);
 	      
-	      GDK_NOTE (DND, g_print ("...CF_UNICODETEXT: %ld bytes\n",
+	      GDK_NOTE (DND, g_print ("... CF_UNICODETEXT: %ld bytes\n",
 				      length));
 
 	      /* Strip out \r */
@@ -441,14 +461,10 @@ gdk_selection_convert (GdkWindow *requestor,
 		  p++;
 		}
 
-	      data = g_utf16_to_utf8 (wcs, wclen, NULL, NULL, &error);
+	      data = g_utf16_to_utf8 (wcs, wclen, NULL, NULL, NULL);
 	      g_free (wcs);
 
-	      if (!data)
-		{
-		  g_error_free (error);
-		}
-	      else
+	      if (data)
 		_gdk_selection_property_store (requestor, target, 8,
 					       data, strlen (data) + 1);
 	      GlobalUnlock (hdata);
@@ -464,7 +480,7 @@ gdk_selection_convert (GdkWindow *requestor,
 	    {
 	      length = GlobalSize (hdata);
 	      
-	      GDK_NOTE (DND, g_print ("...UTF8_STRING: %d bytes: %.10s\n",
+	      GDK_NOTE (DND, g_print ("... UTF8_STRING: %d bytes: %.10s\n",
 				      length, ptr));
 	      
 	      _gdk_selection_property_store (requestor, target, 8,
@@ -488,7 +504,7 @@ gdk_selection_convert (GdkWindow *requestor,
 	    {
 	      length = GlobalSize (hdata);
 	      
-	      GDK_NOTE (DND, g_print ("...CF_TEXT: %ld bytes: %.10s\n",
+	      GDK_NOTE (DND, g_print ("... CF_TEXT: %ld bytes: %.10s\n",
 				       length, ptr));
 	      
 	      if ((hlcid = GetClipboardData (CF_LOCALE)) != NULL)
@@ -499,7 +515,7 @@ gdk_selection_convert (GdkWindow *requestor,
 				     buf, sizeof (buf)))
 		    {
 		      cp = atoi (buf);
-		      GDK_NOTE (DND, g_print ("...CF_LOCALE: %#lx cp:%d\n",
+		      GDK_NOTE (DND, g_print ("... CF_LOCALE: %#lx cp:%d\n",
 					      *lcidptr, cp));
 		    }
 		  GlobalUnlock (hlcid);
@@ -525,12 +541,10 @@ gdk_selection_convert (GdkWindow *requestor,
 		}
 	      g_free (wcs);
 
-	      data = g_utf16_to_utf8 (wcs2, wclen2, NULL, &length, &error);
+	      data = g_utf16_to_utf8 (wcs2, wclen2, NULL, &length, NULL);
 	      g_free (wcs2);
 
-	      if (!data)
-		g_error_free (error);
-	      else
+	      if (data)
 		_gdk_selection_property_store (requestor, target, 8,
 					       data, length + 1);
 	      GlobalUnlock (hdata);
@@ -541,22 +555,42 @@ gdk_selection_convert (GdkWindow *requestor,
 
       API_CALL (CloseClipboard, ());
     }
-  else if (selection == GDK_SELECTION_CLIPBOARD &&
-           target == gdk_atom_intern ("image/bmp", TRUE))
+  else if (selection == GDK_SELECTION_CLIPBOARD && target == _image_bmp)
     {
+      guchar *data;
+
       if (!API_CALL (OpenClipboard, (GDK_WINDOW_HWND (requestor))))
 	return;
-      if ((hdata = GetClipboardData (CF_DIB)) != NULL)
+      if ((hdata = GetClipboardData (_cf_image_bmp)) != NULL)
+	{
+	  /* "image/bmp" is the first choice. */
+	  guchar *ptr;
+
+	  if ((ptr = GlobalLock (hdata)) != NULL)
+	    {
+	      gint length = GlobalSize (hdata);
+      
+	      GDK_NOTE (DND, g_print ("...BITMAP (from \"image/bmp\": %d bytes\n",
+				      length));
+      
+	      _gdk_selection_property_store (requestor, target, 8,
+					     g_memdup (ptr, length), length);
+	      GlobalUnlock (hdata);
+	    }
+	}
+      else if ((hdata = GetClipboardData (CF_DIB)) != NULL)
         {
+	  /* If there's CF_DIB but not "image/bmp", the clipboard
+	   * owner is probably a native Win32 application.
+	   */
           BITMAPINFOHEADER *ptr;
-          guchar *data;
 
           if ((ptr = GlobalLock (hdata)) != NULL)
             {
-              BITMAPFILEHEADER *hdr; /* need to add a file header so gdk-pixbuf can load it */
-	      gint length = GlobalSize (hdata) + sizeof(BITMAPFILEHEADER);
+              BITMAPFILEHEADER *hdr; /* Need to add a file header so gdk-pixbuf can load it */
+	      gint length = GlobalSize (hdata) + sizeof (BITMAPFILEHEADER);
 	      
-	      GDK_NOTE (DND, g_print ("...BITMAP: %d bytes\n", length));
+	      GDK_NOTE (DND, g_print ("... BITMAP (from CF_DIB): %d bytes\n", length));
 	      
               data = g_try_malloc (length);
               if (data)
@@ -564,16 +598,16 @@ gdk_selection_convert (GdkWindow *requestor,
                   hdr = (BITMAPFILEHEADER *)data;
                   hdr->bfType = 0x4d42; /* 0x42 = "B" 0x4d = "M" */
                   /* Compute the size of the entire file. */
-                  hdr->bfSize = (DWORD) (sizeof(BITMAPFILEHEADER)
+                  hdr->bfSize = (DWORD) (sizeof (BITMAPFILEHEADER)
                         + ptr->biSize + ptr->biClrUsed
-			* sizeof(RGBQUAD) + ptr->biSizeImage);
+			* sizeof (RGBQUAD) + ptr->biSizeImage);
                   hdr->bfReserved1 = 0;
                   hdr->bfReserved2 = 0;
                   /* Compute the offset to the array of color indices. */
-                  hdr->bfOffBits = (DWORD) sizeof(BITMAPFILEHEADER)
+                  hdr->bfOffBits = (DWORD) sizeof (BITMAPFILEHEADER)
                         + ptr->biSize + ptr->biClrUsed * sizeof (RGBQUAD);
-                  /* copy the data behind it */
-                  memcpy (data + sizeof(BITMAPFILEHEADER), ptr, length - sizeof(BITMAPFILEHEADER));
+                  /* Copy the data behind it */
+                  memcpy (data + sizeof (BITMAPFILEHEADER), ptr, length - sizeof (BITMAPFILEHEADER));
 	          _gdk_selection_property_store (requestor, target, 8,
 					         data, length);
                 }
@@ -586,22 +620,30 @@ gdk_selection_convert (GdkWindow *requestor,
     }
   else if (selection == GDK_SELECTION_CLIPBOARD)
     {
-      const char *targetname = gdk_atom_name (target);
+      char *target_name;
       UINT fmt = 0;
 
       if (!API_CALL (OpenClipboard, (GDK_WINDOW_HWND (requestor))))
 	return;
-      /* check if its available */
+
+      target_name = gdk_atom_name (target);
+
+      /* Check if it's available. In fact, we can simply call
+       * GetClipboardData (RegisterClipboardFormat (targetname)), but
+       * the global custom format ID space is limited,
+       * (0xC000~0xFFFF), and we better not waste an format ID if we
+       * are just a requestor.
+       */
       for ( ; 0 != (fmt = EnumClipboardFormats (fmt)); )
         {
           char sFormat[80];
 
           if (GetClipboardFormatName (fmt, sFormat, 80) > 0 && 
-              strcmp (sFormat, targetname) == 0)
+              strcmp (sFormat, target_name) == 0)
             {
               if ((hdata = GetClipboardData (fmt)) != NULL)
 	        {
-	          /* simply get it without conversion */
+	          /* Simply get it without conversion */
                   guchar *ptr;
                   gint length;
 
@@ -609,7 +651,7 @@ gdk_selection_convert (GdkWindow *requestor,
                     {
                       length = GlobalSize (hdata);
 	      
-                      GDK_NOTE (DND, g_print ("... %s: %d bytes\n", targetname, length));
+                      GDK_NOTE (DND, g_print ("... %s: %d bytes\n", target_name, length));
 	      
                       _gdk_selection_property_store (requestor, target, 8,
 					             g_memdup (ptr, length), length);
@@ -619,6 +661,7 @@ gdk_selection_convert (GdkWindow *requestor,
                 }
             }
         }
+      g_free (target_name);
       API_CALL (CloseClipboard, ());
     }
   else if (selection == _gdk_win32_dropfiles)
@@ -673,7 +716,8 @@ gdk_selection_property_get (GdkWindow  *requestor,
       return 0;
     }
 
-  *data = g_malloc (prop->length);
+  *data = g_malloc (prop->length + 1);
+  (*data)[prop->length] = '\0';
   if (prop->length > 0)
     memmove (*data, prop->data, prop->length);
 
@@ -713,23 +757,26 @@ gdk_selection_send_notify_for_display (GdkDisplay *display,
                                        GdkAtom     property,
                                        guint32     time)
 {
-  GdkEvent tmp_event;
-  gchar *sel_name, *tgt_name, *prop_name;
-
   g_return_if_fail (display == _gdk_display);
 
-  GDK_NOTE (DND,
-	    (sel_name = gdk_atom_name (selection),
-	     tgt_name = gdk_atom_name (target),
-	     prop_name = gdk_atom_name (property),
-	     g_print ("gdk_selection_send_notify: %#x %#x (%s) %#x (%s) %#x (%s)\n",
-		      requestor,
-		      (guint) selection, sel_name,
-		      (guint) target, tgt_name,
-		      (guint) property, prop_name),
-	     g_free (sel_name),
-	     g_free (tgt_name),
-	     g_free (prop_name)));
+#ifdef G_ENABLE_DEBUG
+  {
+    gchar *sel_name, *tgt_name, *prop_name;
+
+    GDK_NOTE (DND,
+	      (sel_name = gdk_atom_name (selection),
+	       tgt_name = gdk_atom_name (target),
+	       prop_name = gdk_atom_name (property),
+	       g_print ("gdk_selection_send_notify_for_display: %p %#x (%s) %#x (%s) %#x (%s)\n",
+			(gpointer) requestor,
+			(guint) selection, sel_name,
+			(guint) target, tgt_name,
+			(guint) property, prop_name),
+	       g_free (sel_name),
+	       g_free (tgt_name),
+	       g_free (prop_name)));
+  }
+#endif
 }
 
 /* It's hard to say whether implementing this actually is of any use
@@ -744,38 +791,41 @@ gdk_text_property_to_text_list_for_display (GdkDisplay   *display,
 					    gint          length,
 					    gchar      ***list)
 {
-  GError *error = NULL;
-  gchar *enc_name;
   gchar *result;
   const gchar *charset;
-  const gchar *source_charset = NULL;
+  gchar *source_charset;
 
   g_return_val_if_fail (display == _gdk_display, 0);
 
-  GDK_NOTE (DND, (enc_name = gdk_atom_name (encoding),
-		  g_print ("gdk_text_property_to_text_list: %s %d %.20s %d\n",
-			   enc_name, format, text, length),
-		  g_free (enc_name)));
+#ifdef G_ENABLE_DEBUG
+  {
+    gchar *enc_name;
 
+    GDK_NOTE (DND, (enc_name = gdk_atom_name (encoding),
+		    g_print ("gdk_text_property_to_text_list_for_display: %s %d %.20s %d\n",
+			     enc_name, format, text, length),
+		    g_free (enc_name)));
+  }
+#endif
+    
   if (!list)
     return 0;
 
   if (encoding == GDK_TARGET_STRING)
-    source_charset = "ISO-8859-1";
+    source_charset = g_strdup ("ISO-8859-1");
   else if (encoding == _utf8_string)
-    source_charset = "UTF-8";
+    source_charset = g_strdup ("UTF-8");
   else
     source_charset = gdk_atom_name (encoding);
     
   g_get_charset (&charset);
 
   result = g_convert (text, length, charset, source_charset,
-		      NULL, NULL, &error);
+		      NULL, NULL, NULL);
+  g_free (source_charset);
+
   if (!result)
-    {
-      g_error_free (error);
-      return 0;
-    }
+    return 0;
 
   *list = g_new (gchar *, 1);
   **list = result;
@@ -883,7 +933,10 @@ gdk_text_property_to_utf8_list_for_display (GdkDisplay    *display,
     }
   else
     {
-      g_warning ("gdk_text_property_to_utf8_list_for_display: encoding %s not handled\n", gdk_atom_name (encoding));
+      gchar *enc_name = gdk_atom_name (encoding);
+
+      g_warning ("gdk_text_property_to_utf8_list_for_display: encoding %s not handled\n", enc_name);
+      g_free (enc_name);
 
       if (list)
 	*list = NULL;
@@ -968,3 +1021,216 @@ gdk_free_compound_text (guchar *ctext)
    */
   g_return_if_fail (ctext == NULL);
 }
+
+void
+gdk_win32_selection_add_targets (GdkWindow  *owner,
+				 GdkAtom     selection,
+				 gint	     n_targets,
+				 GdkAtom    *targets)
+{
+  HWND hwnd;
+  guint formatid;
+  gint i;
+  GSList *convertable_formats, *format;
+  gboolean has_set_dib = FALSE, has_real_dib = FALSE;
+
+#ifdef G_ENABLE_DEBUG
+  if (_gdk_debug_flags & GDK_DEBUG_DND)
+    {
+      gchar *sel_name = gdk_atom_name (selection);
+      
+      g_print ("gdk_win32_selection_add_targets: %p: %s: ",
+	       owner ? GDK_WINDOW_HWND (owner) : NULL,
+	       sel_name);
+      g_free (sel_name);
+
+      for (i = 0; i < n_targets; i++)
+	{
+	  gchar *tgt_name = gdk_atom_name (targets[i]);
+
+	  g_print ("%s ", tgt_name);
+	  g_free (tgt_name);
+	}
+      g_print ("\n");
+    }
+#endif
+
+  if (selection != GDK_SELECTION_CLIPBOARD)
+    return;
+
+  if (owner != NULL)
+    {
+      if (GDK_WINDOW_DESTROYED (owner))
+	return;
+      hwnd = GDK_WINDOW_HWND (owner);
+    }
+
+  if (!API_CALL (OpenClipboard, (hwnd)))
+    return;
+
+  convertable_formats = gdk_pixbuf_get_formats ();
+  for (i = 0; i < n_targets; ++i)
+    {
+      gchar *target_name;
+
+      if (targets[i] == _utf8_string ||
+	  targets[i] == GDK_TARGET_STRING ||
+	  targets[i] == _text ||
+	  targets[i] == _compound_text ||
+	  targets[i] == _save_targets)
+	continue;
+
+      target_name = gdk_atom_name (targets[i]);
+      if (!(formatid = RegisterClipboardFormat (target_name)))
+	{
+	  WIN32_API_FAILED ("RegisterClipboardFormat");
+	  API_CALL (CloseClipboard, ());
+	  g_free (target_name);
+	  return;
+	}
+      g_hash_table_replace (_format_atom_table, GINT_TO_POINTER (formatid), targets[i]);
+
+      GDK_NOTE (DND, g_print ("... SetClipboardData(%s,NULL)\n",
+			      _gdk_win32_cf_to_string (formatid)));
+      SetClipboardData (formatid, NULL);
+
+      /* We should replace the previous image format associated with
+       * CF_DIB with "image/bmp" if we find "image/bmp", "image/x-bmp"
+       * or "image/x-MS-bmp" is available.
+       */
+      if (!has_real_dib &&
+	  (!strcmp (target_name, "image/bmp") ||
+	   !strcmp (target_name, "image/x-bmp") ||
+	   !strcmp (target_name, "image/x-MS-bmp")))
+	{
+	  g_hash_table_replace (_format_atom_table,
+				GINT_TO_POINTER (CF_DIB),
+				targets[i]);
+	  if (!has_set_dib)
+	    {
+	      GDK_NOTE (DND, g_print ("... SetClipboardData(CF_DIB,NULL)\n"));
+	      SetClipboardData (CF_DIB, NULL);
+	      has_set_dib = TRUE;
+	    }
+	  has_real_dib = TRUE;
+	  g_free (target_name);
+	  continue;
+	}
+
+      for (format = convertable_formats; !has_set_dib && format; format = format->next)
+	{
+	  gchar **mime_types =
+	    gdk_pixbuf_format_get_mime_types ((GdkPixbufFormat *) format->data);
+	  gchar **mime_type;
+
+	  for (mime_type = mime_types; *mime_type; ++mime_type)
+	    {
+	      if (!strcmp (target_name, *mime_type))
+	        {
+		  g_hash_table_replace (_format_atom_table,
+					GINT_TO_POINTER (CF_DIB),
+					targets[i]);
+		  GDK_NOTE (DND, g_print ("... SetClipboardData(CF_DIB,NULL)\n"));
+		  SetClipboardData (CF_DIB, NULL);
+		  has_set_dib = TRUE;
+		  break;
+		}
+	    }
+	    g_strfreev(mime_types);
+	}
+      g_free (target_name);
+    }
+  g_slist_free (convertable_formats);
+
+  API_CALL (CloseClipboard, ());
+}
+
+/* Convert from types such as "image/jpg" or "image/png" to DIB using
+ * gdk-pixbuf so that image copied from GTK+ apps can be pasted in
+ * native apps like mspaint.exe
+ */
+HGLOBAL
+_gdk_win32_selection_convert_to_dib (HGLOBAL  hdata,
+				     GdkAtom  target)
+{
+  GdkPixbufLoader *loader;
+  GdkPixbuf *pixbuf;
+  gchar *target_name;
+  guchar *ptr;
+  gchar *bmp_buf;
+  gsize size;
+  gboolean ok;
+
+  if (!(target_name = gdk_atom_name (target)))
+    {
+      GlobalFree (hdata);
+      return NULL;
+    }
+
+  if (!strcmp (target_name, "image/bmp") ||
+      !strcmp (target_name, "image/x-bmp") ||
+      !strcmp (target_name, "image/x-MS-bmp"))
+    {
+      /* No conversion is needed, just strip the BITMAPFILEHEADER */
+      HGLOBAL hdatanew;
+
+      g_free (target_name);
+      size = GlobalSize (hdata) - 1 - sizeof (BITMAPFILEHEADER);
+      ptr = GlobalLock (hdata);
+      memmove (ptr, ptr + sizeof (BITMAPFILEHEADER), size);
+      GlobalUnlock (hdata);
+      if (!(hdatanew = GlobalReAlloc (hdata, size, 0)))
+	{
+	  WIN32_API_FAILED ("GlobalReAlloc");
+	  GlobalFree (hdata); /* the old hdata is not freed if error */
+	}
+      return hdatanew;
+    }
+
+  /* We actually provide image formats -other than- "image/bmp" etc
+   * and the requestor is either a native Win32 application or a GTK+
+   * client that requested "image/bmp".
+   */
+  if (!(loader = gdk_pixbuf_loader_new_with_mime_type (target_name, NULL)))
+    {
+      GlobalFree (hdata);
+      g_free (target_name);
+      return NULL;
+    }
+  g_free (target_name);
+
+  ptr = GlobalLock (hdata);
+  ok = gdk_pixbuf_loader_write (loader, ptr, GlobalSize (hdata) - 1, NULL) &&
+       gdk_pixbuf_loader_close (loader, NULL);
+
+  GlobalUnlock (hdata);
+  GlobalFree (hdata);
+  hdata = NULL;
+
+  if (ok && (pixbuf = gdk_pixbuf_loader_get_pixbuf (loader)) != NULL)
+    g_object_ref (pixbuf);
+
+  g_object_unref (loader);
+
+  if (ok && gdk_pixbuf_save_to_buffer (pixbuf, &bmp_buf, &size, "bmp", NULL, NULL))
+    {
+      size -= sizeof (BITMAPFILEHEADER);
+      if (!(hdata = GlobalAlloc (GMEM_MOVEABLE, size)))
+	{
+	  WIN32_API_FAILED ("GlobalAlloc");
+	  ok = FALSE;
+	}
+
+      if (ok)
+	{
+	  ptr = GlobalLock (hdata);
+	  memcpy (ptr, bmp_buf + sizeof (BITMAPFILEHEADER), size);
+	  GlobalUnlock (hdata);
+	}
+
+      g_free (bmp_buf);
+      g_object_unref (pixbuf);
+    }
+
+  return hdata;
+}
