@@ -1,15 +1,65 @@
-#ifndef GTK_TEXT_TAG_H
-#define GTK_TEXT_TAG_H
+/* gtktexttag.c - text tag object
+ *
+ * Copyright (c) 1992-1994 The Regents of the University of California.
+ * Copyright (c) 1994-1997 Sun Microsystems, Inc.
+ * Copyright (c) 2000      Red Hat, Inc.
+ * Tk -> Gtk port by Havoc Pennington <hp@redhat.com>
+ *
+ * This software is copyrighted by the Regents of the University of
+ * California, Sun Microsystems, Inc., and other parties.  The
+ * following terms apply to all files associated with the software
+ * unless explicitly disclaimed in individual files.
+ *
+ * The authors hereby grant permission to use, copy, modify,
+ * distribute, and license this software and its documentation for any
+ * purpose, provided that existing copyright notices are retained in
+ * all copies and that this notice is included verbatim in any
+ * distributions. No written agreement, license, or royalty fee is
+ * required for any of the authorized uses.  Modifications to this
+ * software may be copyrighted by their authors and need not follow
+ * the licensing terms described here, provided that the new terms are
+ * clearly indicated on the first page of each file where they apply.
+ *
+ * IN NO EVENT SHALL THE AUTHORS OR DISTRIBUTORS BE LIABLE TO ANY
+ * PARTY FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL
+ * DAMAGES ARISING OUT OF THE USE OF THIS SOFTWARE, ITS DOCUMENTATION,
+ * OR ANY DERIVATIVES THEREOF, EVEN IF THE AUTHORS HAVE BEEN ADVISED
+ * OF THE POSSIBILITY OF SUCH DAMAGE.
+ *
+ * THE AUTHORS AND DISTRIBUTORS SPECIFICALLY DISCLAIM ANY WARRANTIES,
+ * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
+ * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, AND
+ * NON-INFRINGEMENT.  THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS,
+ * AND THE AUTHORS AND DISTRIBUTORS HAVE NO OBLIGATION TO PROVIDE
+ * MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
+ *
+ * GOVERNMENT USE: If you are acquiring this software on behalf of the
+ * U.S. government, the Government shall have only "Restricted Rights"
+ * in the software and related documentation as defined in the Federal
+ * Acquisition Regulations (FARs) in Clause 52.227.19 (c) (2).  If you
+ * are acquiring the software on behalf of the Department of Defense,
+ * the software shall be classified as "Commercial Computer Software"
+ * and the Government shall have only "Restricted Rights" as defined
+ * in Clause 252.227-7013 (c) (1) of DFARs.  Notwithstanding the
+ * foregoing, the authors grant the U.S. Government and others acting
+ * in its behalf permission to use and distribute the software in
+ * accordance with the terms specified in this license.
+ *
+ */
 
-#include <gtk/gtkobject.h>
+#ifndef __GTK_TEXT_TAG_H__
+#define __GTK_TEXT_TAG_H__
+
+#include <glib-object.h>
 #include <gdk/gdk.h>
+#include <gtk/gtkenums.h>
 
-#ifdef __cplusplus
-extern "C" {
-#endif /* __cplusplus */
+/* Not needed, retained for compatibility -Yosh */
+#include <gtk/gtkobject.h>
+
+G_BEGIN_DECLS
 
 typedef struct _GtkTextIter GtkTextIter;
-typedef struct _GtkTextBTreeNode GtkTextBTreeNode;
 typedef struct _GtkTextTagTable GtkTextTagTable;
 
 typedef struct _GtkTextAttributes GtkTextAttributes;
@@ -21,6 +71,8 @@ typedef struct _GtkTextAttributes GtkTextAttributes;
 #define GTK_IS_TEXT_TAG_CLASS(klass) (G_TYPE_CHECK_CLASS_TYPE ((klass), GTK_TYPE_TEXT_TAG))
 #define GTK_TEXT_TAG_GET_CLASS(obj)  (G_TYPE_INSTANCE_GET_CLASS ((obj), GTK_TYPE_TEXT_TAG, GtkTextTagClass))
 
+#define GTK_TYPE_TEXT_ATTRIBUTES     (gtk_text_attributes_get_type ())
+
 typedef struct _GtkTextTag GtkTextTag;
 typedef struct _GtkTextTagClass GtkTextTagClass;
 
@@ -47,7 +99,7 @@ struct _GtkTextTag
    */
 
   GtkTextAttributes *values;
-
+  
   /* Flags for whether a given value is set; if a value is unset, then
    * this tag does not affect it.
    */
@@ -72,12 +124,15 @@ struct _GtkTextTag
   guint invisible_set : 1;
   guint editable_set : 1;
   guint language_set : 1;
+  guint pg_bg_color_set : 1;
+
+  /* Whether these margins accumulate or override */
+  guint accumulative_margin : 1;
+
   guint pad1 : 1;
-  guint pad2 : 1;
-  guint pad3 : 1;
 };
 
-struct _GtkTextTagClass 
+struct _GtkTextTagClass
 {
   GObjectClass parent_class;
 
@@ -85,6 +140,12 @@ struct _GtkTextTagClass
                       GObject           *event_object, /* widget, canvas item, whatever */
                       GdkEvent          *event,        /* the event itself */
                       const GtkTextIter *iter);        /* location of event in buffer */
+
+  /* Padding for future expansion */
+  void (*_gtk_reserved1) (void);
+  void (*_gtk_reserved2) (void);
+  void (*_gtk_reserved3) (void);
+  void (*_gtk_reserved4) (void);
 };
 
 GType        gtk_text_tag_get_type     (void) G_GNUC_CONST;
@@ -105,6 +166,7 @@ typedef struct _GtkTextAppearance GtkTextAppearance;
 
 struct _GtkTextAppearance
 {
+  /*< public >*/
   GdkColor bg_color;
   GdkColor fg_color;
   GdkBitmap *bg_stipple;
@@ -112,7 +174,14 @@ struct _GtkTextAppearance
 
   /* super/subscript rise, can be negative */
   gint rise;
-  
+
+  /*< private >*/
+  /* I'm not sure this can really be used without breaking some things
+   * an app might do :-/
+   */
+  gpointer padding1;
+
+  /*< public >*/
   guint underline : 4;          /* PangoUnderline */
   guint strikethrough : 1;
 
@@ -129,12 +198,20 @@ struct _GtkTextAppearance
    */
   guint inside_selection : 1;
   guint is_text : 1;
+
+  /*< private >*/
+  guint pad1 : 1;
+  guint pad2 : 1;
+  guint pad3 : 1;
+  guint pad4 : 1;
 };
 
 struct _GtkTextAttributes
 {
+  /*< private >*/
   guint refcount;
 
+  /*< public >*/
   GtkTextAppearance appearance;
 
   GtkJustification justification;
@@ -166,6 +243,10 @@ struct _GtkTextAttributes
 
   PangoLanguage *language;
 
+  /*< private >*/
+  GdkColor *pg_bg_color;
+
+  /*< public >*/
   /* hide the text  */
   guint invisible : 1;
 
@@ -180,6 +261,7 @@ struct _GtkTextAttributes
   /* colors are allocated etc. */
   guint realized : 1;
 
+  /*< private >*/
   guint pad1 : 1;
   guint pad2 : 1;
   guint pad3 : 1;
@@ -191,12 +273,12 @@ GtkTextAttributes* gtk_text_attributes_copy        (GtkTextAttributes *src);
 void               gtk_text_attributes_copy_values (GtkTextAttributes *src,
                                                     GtkTextAttributes *dest);
 void               gtk_text_attributes_unref       (GtkTextAttributes *values);
-void               gtk_text_attributes_ref         (GtkTextAttributes *values);
+GtkTextAttributes *gtk_text_attributes_ref         (GtkTextAttributes *values);
+
+GType              gtk_text_attributes_get_type    (void) G_GNUC_CONST;
 
 
-#ifdef __cplusplus
-}
-#endif /* __cplusplus */
+G_END_DECLS
 
 #endif
 
