@@ -31,6 +31,7 @@
 #include <config.h>
 
 #include <string.h>
+#include "gtkbuildable.h"
 #include "gtkintl.h"
 #include "gtkmarshalers.h"
 #include "gtkmenu.h"
@@ -107,8 +108,6 @@ struct _NodeUIReference
   GQuark action_quark;
 };
 
-static void        gtk_ui_manager_class_init      (GtkUIManagerClass *class);
-static void        gtk_ui_manager_init            (GtkUIManager      *self);
 static void        gtk_ui_manager_finalize        (GObject           *object);
 static void        gtk_ui_manager_set_property    (GObject           *object,
                                                    guint              prop_id,
@@ -144,6 +143,28 @@ static void        node_prepend_ui_reference      (GNode             *node,
 static void        node_remove_ui_reference       (GNode             *node,
                                                    guint              merge_id);
 
+/* GtkBuildable */
+static void gtk_ui_manager_buildable_init      (GtkBuildableIface *iface);
+static void gtk_ui_manager_buildable_add_child (GtkBuildable  *buildable,
+						GtkBuilder    *builder,
+						GObject       *child,
+						const gchar   *type);
+static GObject* gtk_ui_manager_buildable_construct_child (GtkBuildable *buildable,
+							  GtkBuilder   *builder,
+							  const gchar  *name);
+static gboolean gtk_ui_manager_buildable_custom_tag_start (GtkBuildable  *buildable,
+							   GtkBuilder    *builder,
+							   GObject       *child,
+							   const gchar   *tagname,
+							   GMarkupParser *parser,
+							   gpointer      *data);
+static void     gtk_ui_manager_buildable_custom_tag_end (GtkBuildable 	 *buildable,
+							 GtkBuilder   	 *builder,
+							 GObject      	 *child,
+							 const gchar  	 *tagname,
+							 gpointer     	 *data);
+
+
 
 enum 
 {
@@ -163,52 +184,19 @@ enum
   PROP_UI
 };
 
-static GObjectClass *parent_class = NULL;
 static guint ui_manager_signals[LAST_SIGNAL] = { 0 };
 
-static GMemChunk *merge_node_chunk = NULL;
-
-GType
-gtk_ui_manager_get_type (void)
-{
-  static GtkType type = 0;
-
-  if (!type)
-    {
-      static const GTypeInfo type_info =
-      {
-        sizeof (GtkUIManagerClass),
-        (GBaseInitFunc) NULL,
-        (GBaseFinalizeFunc) NULL,
-        (GClassInitFunc) gtk_ui_manager_class_init,
-        (GClassFinalizeFunc) NULL,
-        NULL,
-        
-        sizeof (GtkUIManager),
-        0, /* n_preallocs */
-        (GInstanceInitFunc) gtk_ui_manager_init,
-      };
-
-      type = g_type_register_static (G_TYPE_OBJECT,
-				     "GtkUIManager",
-				     &type_info, 0);
-    }
-  return type;
-}
+G_DEFINE_TYPE_WITH_CODE (GtkUIManager, gtk_ui_manager, G_TYPE_OBJECT,
+			 G_IMPLEMENT_INTERFACE (GTK_TYPE_BUILDABLE,
+						gtk_ui_manager_buildable_init))
 
 static void
 gtk_ui_manager_class_init (GtkUIManagerClass *klass)
 {
   GObjectClass *gobject_class;
 
-  parent_class = g_type_class_peek_parent (klass);
-
   gobject_class = G_OBJECT_CLASS (klass);
 
-  if (!merge_node_chunk)
-    merge_node_chunk = g_mem_chunk_create (Node, 64,
-					   G_ALLOC_AND_FREE);
-
   gobject_class->finalize = gtk_ui_manager_finalize;
   gobject_class->set_property = gtk_ui_manager_set_property;
   gobject_class->get_property = gtk_ui_manager_get_property;
@@ -239,7 +227,7 @@ gtk_ui_manager_class_init (GtkUIManagerClass *klass)
 				   g_param_spec_string ("ui",
  							P_("Merged UI definition"),
 							P_("An XML string describing the merged UI"),
-							"<ui>\n</ui>",
+							"<ui>\n</ui>\n",
 							GTK_PARAM_READABLE));
 
 
@@ -255,7 +243,7 @@ gtk_ui_manager_class_init (GtkUIManagerClass *klass)
    * Since: 2.4
    */
   ui_manager_signals[ADD_WIDGET] =
-    g_signal_new ("add_widget",
+    g_signal_new (I_("add_widget"),
 		  G_OBJECT_CLASS_TYPE (klass),
 		  G_SIGNAL_RUN_FIRST | G_SIGNAL_NO_RECURSE,
 		  G_STRUCT_OFFSET (GtkUIManagerClass, add_widget),
@@ -274,7 +262,7 @@ gtk_ui_manager_class_init (GtkUIManagerClass *klass)
    * Since: 2.4
    */
   ui_manager_signals[ACTIONS_CHANGED] =
-    g_signal_new ("actions_changed",
+    g_signal_new (I_("actions_changed"),
 		  G_OBJECT_CLASS_TYPE (klass),
 		  G_SIGNAL_RUN_FIRST | G_SIGNAL_NO_RECURSE,
 		  G_STRUCT_OFFSET (GtkUIManagerClass, actions_changed),  
@@ -298,7 +286,7 @@ gtk_ui_manager_class_init (GtkUIManagerClass *klass)
    * Since: 2.4
    */
   ui_manager_signals[CONNECT_PROXY] =
-    g_signal_new ("connect_proxy",
+    g_signal_new (I_("connect_proxy"),
 		  G_OBJECT_CLASS_TYPE (klass),
 		  G_SIGNAL_RUN_FIRST | G_SIGNAL_NO_RECURSE,
 		  G_STRUCT_OFFSET (GtkUIManagerClass, connect_proxy),
@@ -320,7 +308,7 @@ gtk_ui_manager_class_init (GtkUIManagerClass *klass)
    * Since: 2.4
    */
   ui_manager_signals[DISCONNECT_PROXY] =
-    g_signal_new ("disconnect_proxy",
+    g_signal_new (I_("disconnect_proxy"),
 		  G_OBJECT_CLASS_TYPE (klass),
 		  G_SIGNAL_RUN_FIRST | G_SIGNAL_NO_RECURSE,
 		  G_STRUCT_OFFSET (GtkUIManagerClass, disconnect_proxy),
@@ -344,7 +332,7 @@ gtk_ui_manager_class_init (GtkUIManagerClass *klass)
    * Since: 2.4
    */
   ui_manager_signals[PRE_ACTIVATE] =
-    g_signal_new ("pre_activate",
+    g_signal_new (I_("pre_activate"),
 		  G_OBJECT_CLASS_TYPE (klass),
 		  G_SIGNAL_RUN_FIRST | G_SIGNAL_NO_RECURSE,
 		  G_STRUCT_OFFSET (GtkUIManagerClass, pre_activate),
@@ -367,7 +355,7 @@ gtk_ui_manager_class_init (GtkUIManagerClass *klass)
    * Since: 2.4
    */
   ui_manager_signals[POST_ACTIVATE] =
-    g_signal_new ("post_activate",
+    g_signal_new (I_("post_activate"),
 		  G_OBJECT_CLASS_TYPE (klass),
 		  G_SIGNAL_RUN_FIRST | G_SIGNAL_NO_RECURSE,
 		  G_STRUCT_OFFSET (GtkUIManagerClass, post_activate),
@@ -434,7 +422,82 @@ gtk_ui_manager_finalize (GObject *object)
   g_object_unref (self->private_data->accel_group);
   self->private_data->accel_group = NULL;
 
-  G_OBJECT_CLASS (parent_class)->finalize (object);
+  G_OBJECT_CLASS (gtk_ui_manager_parent_class)->finalize (object);
+}
+
+static void
+gtk_ui_manager_buildable_init (GtkBuildableIface *iface)
+{
+  iface->add_child = gtk_ui_manager_buildable_add_child;
+  iface->construct_child = gtk_ui_manager_buildable_construct_child;
+  iface->custom_tag_start = gtk_ui_manager_buildable_custom_tag_start;
+  iface->custom_tag_end = gtk_ui_manager_buildable_custom_tag_end;
+}
+
+static void
+gtk_ui_manager_buildable_add_child (GtkBuildable  *buildable,
+				    GtkBuilder    *builder,
+				    GObject       *child,
+				    const gchar   *type)
+{
+  GtkUIManager *self = GTK_UI_MANAGER (buildable);
+  guint pos;
+
+  g_return_if_fail (GTK_IS_ACTION_GROUP (child));
+
+  pos = g_list_length (self->private_data->action_groups);
+
+  g_object_ref (child);
+  gtk_ui_manager_insert_action_group (self,
+				      GTK_ACTION_GROUP (child),
+				      pos);
+}
+
+static void
+child_hierarchy_changed_cb (GtkWidget *widget,
+			    GtkWidget *unused,
+			    GtkUIManager *uimgr)
+{
+  GtkWidget *toplevel;
+  GtkAccelGroup *group;
+  GSList *groups;
+
+  toplevel = gtk_widget_get_toplevel (widget);
+  if (!toplevel || !GTK_IS_WINDOW (toplevel))
+    return;
+  
+  group = gtk_ui_manager_get_accel_group (uimgr);
+  groups = gtk_accel_groups_from_object (toplevel);
+  if (g_slist_find (groups, group) == NULL)
+    gtk_window_add_accel_group (GTK_WINDOW (toplevel), group);
+
+  g_signal_handlers_disconnect_by_func (widget,
+					child_hierarchy_changed_cb,
+					uimgr);
+}
+
+static GObject *
+gtk_ui_manager_buildable_construct_child (GtkBuildable *buildable,
+					  GtkBuilder   *builder,
+					  const gchar  *id)
+{
+  GtkWidget *widget;
+  char *name;
+
+  name = g_strdup_printf ("ui/%s", id);
+  widget = gtk_ui_manager_get_widget (GTK_UI_MANAGER (buildable), name);
+  if (!widget)
+    {
+      g_error ("Unknown ui manager child: %s\n", name);
+      g_free (name);
+      return NULL;
+    }
+  g_free (name);
+
+  g_signal_connect (widget, "hierarchy-changed",
+		    G_CALLBACK (child_hierarchy_changed_cb),
+		    GTK_UI_MANAGER (buildable));
+  return g_object_ref (widget);
 }
 
 static void
@@ -470,7 +533,7 @@ gtk_ui_manager_get_property (GObject         *object,
       g_value_set_boolean (value, self->private_data->add_tearoffs);
       break;
     case PROP_UI:
-      g_value_set_string (value, gtk_ui_manager_get_ui (self));
+      g_value_take_string (value, gtk_ui_manager_get_ui (self));
       break;
     default:
       G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
@@ -896,7 +959,7 @@ get_child_node (GtkUIManager *self,
 	{
 	  Node *mnode;
 	  
-	  mnode = g_chunk_new0 (Node, merge_node_chunk);
+	  mnode = g_slice_new0 (Node);
 	  mnode->type = node_type;
 	  mnode->name = g_strndup (childname, childname_length);
 
@@ -936,7 +999,7 @@ get_child_node (GtkUIManager *self,
 	{
 	  Node *mnode;
 
-	  mnode = g_chunk_new0 (Node, merge_node_chunk);
+	  mnode = g_slice_new0 (Node);
 	  mnode->type = node_type;
 	  mnode->name = g_strndup (childname, childname_length);
 	  mnode->dirty = TRUE;
@@ -989,12 +1052,18 @@ get_node (GtkUIManager *self,
   return node;
 }
 
+static void
+node_ui_reference_free (gpointer data)
+{
+  g_slice_free (NodeUIReference, data);
+}
+
 static gboolean
 free_node (GNode *node)
 {
   Node *info = NODE_INFO (node);
   
-  g_list_foreach (info->uifiles, (GFunc) g_free, NULL);
+  g_list_foreach (info->uifiles, (GFunc) node_ui_reference_free, NULL);
   g_list_free (info->uifiles);
 
   if (info->action)
@@ -1004,7 +1073,7 @@ free_node (GNode *node)
   if (info->extra)
     g_object_unref (info->extra);
   g_free (info->name);
-  g_chunk_free (info, merge_node_chunk);
+  g_slice_free (Node, info);
 
   return FALSE;
 }
@@ -1041,7 +1110,7 @@ node_prepend_ui_reference (GNode  *gnode,
     reference = node->uifiles->data;
   else
     {
-      reference = g_new (NodeUIReference, 1);
+      reference = g_slice_new (NodeUIReference);
       node->uifiles = g_list_prepend (node->uifiles, reference);
     }
 
@@ -1067,7 +1136,7 @@ node_remove_ui_reference (GNode  *gnode,
 	  if (p == node->uifiles)
 	    mark_node_dirty (gnode);
 	  node->uifiles = g_list_delete_link (node->uifiles, p);
-	  g_free (reference);
+	  g_slice_free (NodeUIReference, reference);
 
 	  break;
 	}
@@ -1308,7 +1377,7 @@ start_element_handler (GMarkupParseContext *context,
 	  if (!strcmp (node_name, "separator"))
 	    {
 	      node_name = NULL;
-	      length = -1;
+	      length = 0;
 	    }
 	  else
 	    length = strlen (node_name);
@@ -1467,7 +1536,7 @@ text_handler (GMarkupParseContext *context,
 }
 
 
-static GMarkupParser ui_parser = {
+static const GMarkupParser ui_parser = {
   start_element_handler,
   end_element_handler,
   text_handler,
@@ -1510,7 +1579,7 @@ add_ui_from_string (GtkUIManager *self,
 
   queue_update (self);
 
-  g_object_notify (G_OBJECT (self), "ui");      
+  g_object_notify (G_OBJECT (self), "ui");
 
   return ctx.merge_id;
 
@@ -1640,7 +1709,7 @@ gtk_ui_manager_add_ui (GtkUIManager        *self,
 
   g_return_if_fail (GTK_IS_UI_MANAGER (self));  
   g_return_if_fail (merge_id > 0);
-  g_return_if_fail (name != NULL);
+  g_return_if_fail (name != NULL || type == GTK_UI_MANAGER_SEPARATOR);
 
   node = get_node (self, path, NODE_TYPE_UNDECIDED, FALSE);
   sibling = NULL;
@@ -1741,7 +1810,7 @@ gtk_ui_manager_add_ui (GtkUIManager        *self,
     }
    
   child = get_child_node (self, node, sibling,
-			  name, strlen (name),
+			  name, name ? strlen (name) : 0,
 			  node_type, TRUE, top);
 
   if (action != NULL)
@@ -1781,6 +1850,8 @@ void
 gtk_ui_manager_remove_ui (GtkUIManager *self, 
 			  guint         merge_id)
 {
+  g_return_if_fail (GTK_IS_UI_MANAGER (self));
+
   g_node_traverse (self->private_data->root_node, 
 		   G_POST_ORDER, G_TRAVERSE_ALL, -1,
 		   remove_ui, GUINT_TO_POINTER (merge_id));
@@ -1823,15 +1894,15 @@ find_menu_position (GNode      *node,
 		    gint       *pos_p)
 {
   GtkWidget *menushell;
-  gint pos;
+  gint pos = 0;
 
   g_return_val_if_fail (node != NULL, FALSE);
   g_return_val_if_fail (NODE_INFO (node)->type == NODE_TYPE_MENU ||
-			NODE_INFO (node)->type == NODE_TYPE_POPUP ||
-			NODE_INFO (node)->type == NODE_TYPE_MENU_PLACEHOLDER ||
-			NODE_INFO (node)->type == NODE_TYPE_MENUITEM ||
-			NODE_INFO (node)->type == NODE_TYPE_SEPARATOR,
-			FALSE);
+		        NODE_INFO (node)->type == NODE_TYPE_POPUP ||
+		        NODE_INFO (node)->type == NODE_TYPE_MENU_PLACEHOLDER ||
+		        NODE_INFO (node)->type == NODE_TYPE_MENUITEM ||
+		        NODE_INFO (node)->type == NODE_TYPE_SEPARATOR,
+                        FALSE);
 
   /* first sibling -- look at parent */
   if (node->prev == NULL)
@@ -1865,8 +1936,8 @@ find_menu_position (GNode      *node,
 			      NODE_INFO (parent)->proxy) + 1;
 	  break;
 	default:
-	  g_warning("%s: bad parent node type %d", G_STRLOC,
-		    NODE_INFO (parent)->type);
+	  g_warning ("%s: bad parent node type %d", G_STRLOC,
+		     NODE_INFO (parent)->type);
 	  return FALSE;
 	}
     }
@@ -1881,9 +1952,12 @@ find_menu_position (GNode      *node,
       else
 	prev_child = NODE_INFO (sibling)->proxy;
 
-      g_return_val_if_fail (GTK_IS_WIDGET (prev_child), FALSE);
+      if (!GTK_IS_WIDGET (prev_child))
+        return FALSE;
+
       menushell = gtk_widget_get_parent (prev_child);
-      g_return_val_if_fail (GTK_IS_MENU_SHELL (menushell), FALSE);
+      if (!GTK_IS_MENU_SHELL (menushell))
+        return FALSE;
 
       pos = g_list_index (GTK_MENU_SHELL (menushell)->children, prev_child) + 1;
     }
@@ -1906,11 +1980,11 @@ find_toolbar_position (GNode      *node,
 
   g_return_val_if_fail (node != NULL, FALSE);
   g_return_val_if_fail (NODE_INFO (node)->type == NODE_TYPE_TOOLBAR ||
-			NODE_INFO (node)->type == NODE_TYPE_TOOLBAR_PLACEHOLDER ||
-			NODE_INFO (node)->type == NODE_TYPE_TOOLITEM ||
-			NODE_INFO (node)->type == NODE_TYPE_SEPARATOR,
-			FALSE);
-
+		        NODE_INFO (node)->type == NODE_TYPE_TOOLBAR_PLACEHOLDER ||
+		        NODE_INFO (node)->type == NODE_TYPE_TOOLITEM ||
+		        NODE_INFO (node)->type == NODE_TYPE_SEPARATOR,
+                        FALSE);
+  
   /* first sibling -- look at parent */
   if (node->prev == NULL)
     {
@@ -1946,14 +2020,17 @@ find_toolbar_position (GNode      *node,
       else
 	prev_child = NODE_INFO (sibling)->proxy;
 
-      g_return_val_if_fail (GTK_IS_WIDGET (prev_child), FALSE);
+      if (!GTK_IS_WIDGET (prev_child))
+        return FALSE;
+
       toolbar = gtk_widget_get_parent (prev_child);
-      g_return_val_if_fail (GTK_IS_TOOLBAR (toolbar), FALSE);
+      if (!GTK_IS_TOOLBAR (toolbar))
+        return FALSE;
 
       pos = gtk_toolbar_get_item_index (GTK_TOOLBAR (toolbar),
 					GTK_TOOL_ITEM (prev_child)) + 1;
     }
-
+  
   if (toolbar_p)
     *toolbar_p = toolbar;
   if (pos_p)
@@ -1978,6 +2055,7 @@ gboolean
 _gtk_menu_is_empty (GtkWidget *menu)
 {
   GList *children, *cur;
+  gboolean result = TRUE;
 
   g_return_val_if_fail (menu == NULL || GTK_IS_MENU (menu), TRUE);
 
@@ -1993,13 +2071,16 @@ _gtk_menu_is_empty (GtkWidget *menu)
 	{
 	  if (!GTK_IS_TEAROFF_MENU_ITEM (cur->data) &&
 	      !g_object_get_data (cur->data, "gtk-empty-menu-item"))
-	    return FALSE;
+            {
+	      result = FALSE;
+              break;
+            }
 	}
       cur = cur->next;
     }
   g_list_free (children);
 
-  return TRUE;
+  return result;
 }
 
 enum {
@@ -2008,10 +2089,6 @@ enum {
   SEPARATOR_MODE_HIDDEN
 };
 
-void _gtk_action_sync_menu_visible (GtkAction *action,
-				    GtkWidget *proxy,
-				    gboolean   empty);
-
 static void
 update_smart_separators (GtkWidget *proxy)
 {
@@ -2028,7 +2105,6 @@ update_smart_separators (GtkWidget *proxy)
       gboolean empty;
       GList *children, *cur, *last;
       GtkWidget *filler;
-      gint i;
 
       children = gtk_container_get_children (GTK_CONTAINER (parent));
       
@@ -2037,7 +2113,6 @@ update_smart_separators (GtkWidget *proxy)
       empty = TRUE;
       filler = NULL;
 
-      i = 0;
       cur = children;
       while (cur) 
 	{
@@ -2172,9 +2247,9 @@ update_node (GtkUIManager *self,
       info->type != NODE_TYPE_MENU_PLACEHOLDER &&
       info->type != NODE_TYPE_TOOLBAR_PLACEHOLDER)
     {
-      g_warning ("%s: missing action", info->name);
+      g_warning ("%s: missing action %s", info->name, action_name);
       
-      goto recurse_children;
+      return;
     }
   
   if (action)
@@ -2214,8 +2289,7 @@ update_node (GtkUIManager *self,
       if (info->proxy == NULL)
 	{
 	  info->proxy = gtk_menu_bar_new ();
-	  g_object_ref (info->proxy);
-	  gtk_object_sink (GTK_OBJECT (info->proxy));
+	  g_object_ref_sink (info->proxy);
 	  gtk_widget_set_name (info->proxy, info->name);
 	  gtk_widget_show (info->proxy);
 	  g_signal_emit (self, ui_manager_signals[ADD_WIDGET], 0, info->proxy);
@@ -2225,16 +2299,16 @@ update_node (GtkUIManager *self,
       if (info->proxy == NULL) 
 	{
 	  info->proxy = gtk_menu_new ();
-	  g_object_ref (info->proxy);
-	  gtk_object_sink (GTK_OBJECT (info->proxy));
+	  g_object_ref_sink (info->proxy);
 	}
       gtk_widget_set_name (info->proxy, info->name);
       break;
     case NODE_TYPE_MENU:
       {
 	GtkWidget *prev_submenu = NULL;
-	GtkWidget *menu;
+	GtkWidget *menu = NULL;
 	GList *siblings;
+
 	/* remove the proxy if it is of the wrong type ... */
 	if (info->proxy &&  
 	    G_OBJECT_TYPE (info->proxy) != GTK_ACTION_GET_CLASS (action)->menu_item_type)
@@ -2255,51 +2329,62 @@ update_node (GtkUIManager *self,
 	    g_object_unref (info->proxy);
 	    info->proxy = NULL;
 	  }
+
 	/* create proxy if needed ... */
 	if (info->proxy == NULL)
 	  {
-	    GtkWidget *menushell;
-	    gint pos;
-	    
-	    if (NODE_INFO (node->parent)->type == NODE_TYPE_TOOLITEM ||
-		find_menu_position (node, &menushell, &pos))
-	      {
-		GtkWidget *tearoff;
-		GtkWidget *filler;
-		
-		menu = gtk_menu_new ();
-		gtk_widget_set_name (menu, info->name);
-		tearoff = gtk_tearoff_menu_item_new ();
-		gtk_widget_set_no_show_all (tearoff, TRUE);
-		gtk_menu_shell_append (GTK_MENU_SHELL (menu), tearoff);
-		filler = gtk_menu_item_new_with_label (_("Empty"));
-		g_object_set_data (G_OBJECT (filler),
-				   "gtk-empty-menu-item",
-				   GINT_TO_POINTER (TRUE));
-		gtk_widget_set_sensitive (filler, FALSE);
-		gtk_widget_set_no_show_all (filler, TRUE);
-		gtk_menu_shell_append (GTK_MENU_SHELL (menu), filler);
+            /* ... if the action already provides a menu, then use
+             * that menu instead of creating an empty one
+             */
+            if ((NODE_INFO (node->parent)->type == NODE_TYPE_TOOLITEM ||
+                 NODE_INFO (node->parent)->type == NODE_TYPE_MENUITEM) &&
+                GTK_ACTION_GET_CLASS (action)->create_menu)
+              {
+                menu = gtk_action_create_menu (action);
+              }
 
-		if (NODE_INFO (node->parent)->type == NODE_TYPE_TOOLITEM)
-		  {
-		    info->proxy = menu;
-		    g_object_ref (info->proxy);
-		    gtk_object_sink (GTK_OBJECT (info->proxy));
-		    gtk_menu_tool_button_set_menu (GTK_MENU_TOOL_BUTTON (NODE_INFO (node->parent)->proxy),
-						   menu);
-		  }
-		else
-		  {
-		    info->proxy = gtk_action_create_menu_item (action);
-		    g_object_ref (info->proxy);
-		    gtk_object_sink (GTK_OBJECT (info->proxy));
-		    g_signal_connect (info->proxy, "notify::visible",
-				      G_CALLBACK (update_smart_separators), NULL);
-		    gtk_widget_set_name (info->proxy, info->name);
-		    
-		    gtk_menu_item_set_submenu (GTK_MENU_ITEM (info->proxy), menu);
-		    gtk_menu_shell_insert (GTK_MENU_SHELL (menushell), info->proxy, pos);
-		  }
+            if (!menu)
+              {
+                GtkWidget *tearoff;
+                GtkWidget *filler;
+	    
+                menu = gtk_menu_new ();
+                gtk_widget_set_name (menu, info->name);
+                tearoff = gtk_tearoff_menu_item_new ();
+                gtk_widget_set_no_show_all (tearoff, TRUE);
+                gtk_menu_shell_append (GTK_MENU_SHELL (menu), tearoff);
+                filler = gtk_menu_item_new_with_label (_("Empty"));
+                g_object_set_data (G_OBJECT (filler),
+                                   I_("gtk-empty-menu-item"),
+                                   GINT_TO_POINTER (TRUE));
+                gtk_widget_set_sensitive (filler, FALSE);
+                gtk_widget_set_no_show_all (filler, TRUE);
+                gtk_menu_shell_append (GTK_MENU_SHELL (menu), filler);
+              }
+	    
+            if (NODE_INFO (node->parent)->type == NODE_TYPE_TOOLITEM)
+	      {
+		info->proxy = menu;
+		g_object_ref_sink (info->proxy);
+		gtk_menu_tool_button_set_menu (GTK_MENU_TOOL_BUTTON (NODE_INFO (node->parent)->proxy),
+					       menu);
+	      }
+	    else
+	      {
+		GtkWidget *menushell;
+		gint pos;
+		
+		if (find_menu_position (node, &menushell, &pos))
+                  {
+		     info->proxy = gtk_action_create_menu_item (action);
+		     g_object_ref_sink (info->proxy);
+		     g_signal_connect (info->proxy, "notify::visible",
+		   		       G_CALLBACK (update_smart_separators), NULL);
+		     gtk_widget_set_name (info->proxy, info->name);
+		
+		     gtk_menu_item_set_submenu (GTK_MENU_ITEM (info->proxy), menu);
+		     gtk_menu_shell_insert (GTK_MENU_SHELL (menushell), info->proxy, pos);
+                 }
 	      }
 	  }
 	else
@@ -2316,6 +2401,7 @@ update_node (GtkUIManager *self,
 	  menu = info->proxy;
 	else
 	  menu = gtk_menu_item_get_submenu (GTK_MENU_ITEM (info->proxy));
+
 	siblings = gtk_container_get_children (GTK_CONTAINER (menu));
 	if (siblings != NULL && GTK_IS_TEAROFF_MENU_ITEM (siblings->data))
 	  {
@@ -2336,8 +2422,7 @@ update_node (GtkUIManager *self,
       if (info->proxy == NULL)
 	{
 	  info->proxy = gtk_toolbar_new ();
-	  g_object_ref (info->proxy);
-	  gtk_object_sink (GTK_OBJECT (info->proxy));
+	  g_object_ref_sink (info->proxy);
 	  gtk_widget_set_name (info->proxy, info->name);
 	  gtk_widget_show (info->proxy);
 	  g_signal_emit (self, ui_manager_signals[ADD_WIDGET], 0, info->proxy);
@@ -2369,27 +2454,25 @@ update_node (GtkUIManager *self,
 	  gint pos;
 	  
 	  if (find_menu_position (node, &menushell, &pos))
-	    {
+            {
 	      info->proxy = gtk_separator_menu_item_new ();
-	      g_object_ref (info->proxy);
-	      gtk_object_sink (GTK_OBJECT (info->proxy));
+	      g_object_ref_sink (info->proxy);
 	      g_object_set_data (G_OBJECT (info->proxy),
-				 "gtk-separator-mode",
-				 GINT_TO_POINTER (SEPARATOR_MODE_HIDDEN));
+	  		         I_("gtk-separator-mode"),
+			         GINT_TO_POINTER (SEPARATOR_MODE_HIDDEN));
 	      gtk_widget_set_no_show_all (info->proxy, TRUE);
 	      gtk_menu_shell_insert (GTK_MENU_SHELL (menushell),
-				     NODE_INFO (node)->proxy, pos);
-	      
+	 			     NODE_INFO (node)->proxy, pos);
+	  
 	      info->extra = gtk_separator_menu_item_new ();
-	      g_object_ref (info->extra);
-	      gtk_object_sink (GTK_OBJECT (info->extra));
+	      g_object_ref_sink (info->extra);
 	      g_object_set_data (G_OBJECT (info->extra),
-				 "gtk-separator-mode",
-				 GINT_TO_POINTER (SEPARATOR_MODE_HIDDEN));
+			         I_("gtk-separator-mode"),
+			         GINT_TO_POINTER (SEPARATOR_MODE_HIDDEN));
 	      gtk_widget_set_no_show_all (info->extra, TRUE);
 	      gtk_menu_shell_insert (GTK_MENU_SHELL (menushell),
-				     NODE_INFO (node)->extra, pos+1);
-	    }
+				     NODE_INFO (node)->extra, pos + 1);
+            }
 	}
       break;
     case NODE_TYPE_TOOLBAR_PLACEHOLDER:
@@ -2416,31 +2499,28 @@ update_node (GtkUIManager *self,
 	{
 	  GtkWidget *toolbar;
 	  gint pos;
+	  GtkToolItem *item;    
 	  
 	  if (find_toolbar_position (node, &toolbar, &pos))
-	    {
-	      GtkToolItem *item;
-	      
+            {
 	      item = gtk_separator_tool_item_new ();
 	      gtk_toolbar_insert (GTK_TOOLBAR (toolbar), item, pos);
 	      info->proxy = GTK_WIDGET (item);
-	      g_object_ref (info->proxy);
-	      gtk_object_sink (GTK_OBJECT (info->proxy));
+	      g_object_ref_sink (info->proxy);
 	      g_object_set_data (G_OBJECT (info->proxy),
-				 "gtk-separator-mode",
-				 GINT_TO_POINTER (SEPARATOR_MODE_HIDDEN));
+			         I_("gtk-separator-mode"),
+			         GINT_TO_POINTER (SEPARATOR_MODE_HIDDEN));
 	      gtk_widget_set_no_show_all (info->proxy, TRUE);
-	      
+	  
 	      item = gtk_separator_tool_item_new ();
 	      gtk_toolbar_insert (GTK_TOOLBAR (toolbar), item, pos+1);
 	      info->extra = GTK_WIDGET (item);
-	      g_object_ref (info->extra);
-	      gtk_object_sink (GTK_OBJECT (info->extra));
+	      g_object_ref_sink (info->extra);
 	      g_object_set_data (G_OBJECT (info->extra),
-				 "gtk-separator-mode",
-				 GINT_TO_POINTER (SEPARATOR_MODE_HIDDEN));
+			         I_("gtk-separator-mode"),
+			         GINT_TO_POINTER (SEPARATOR_MODE_HIDDEN));
 	      gtk_widget_set_no_show_all (info->extra, TRUE);
-	    }
+            }
 	}
       break;
     case NODE_TYPE_MENUITEM:
@@ -2464,15 +2544,14 @@ update_node (GtkUIManager *self,
 	  gint pos;
 	  
 	  if (find_menu_position (node, &menushell, &pos))
-	    {
+            {
 	      info->proxy = gtk_action_create_menu_item (action);
-	      g_object_ref (info->proxy);
-	      gtk_object_sink (GTK_OBJECT (info->proxy));
+	      g_object_ref_sink (info->proxy);
 	      gtk_widget_set_name (info->proxy, info->name);
-	      
+	  
 	      gtk_menu_shell_insert (GTK_MENU_SHELL (menushell),
 				     info->proxy, pos);
-	    }
+           }
 	}
       else
 	{
@@ -2482,16 +2561,18 @@ update_node (GtkUIManager *self,
 	  gtk_menu_item_set_submenu (GTK_MENU_ITEM (info->proxy), NULL);
 	  gtk_action_connect_proxy (action, info->proxy);
 	}
-      g_signal_connect (info->proxy, "notify::visible",
-			G_CALLBACK (update_smart_separators), NULL);
-      if (in_popup) 
-	{
-	  /* don't show accels in popups */
-	  GtkWidget *label = GTK_BIN (info->proxy)->child;
-	  g_object_set (label,
-			"accel_closure", NULL,
-			NULL);
-	}
+
+      if (info->proxy)
+        {
+          g_signal_connect (info->proxy, "notify::visible",
+			    G_CALLBACK (update_smart_separators), NULL);
+          if (in_popup) 
+	    {
+	      /* don't show accels in popups */
+	      GtkWidget *label = GTK_BIN (info->proxy)->child;
+	      g_object_set (label, "accel-closure", NULL, NULL);
+	    }
+        }
       
       break;
     case NODE_TYPE_TOOLITEM:
@@ -2515,15 +2596,14 @@ update_node (GtkUIManager *self,
 	  gint pos;
 	  
 	  if (find_toolbar_position (node, &toolbar, &pos))
-	    {
+            {
 	      info->proxy = gtk_action_create_tool_item (action);
-	      g_object_ref (info->proxy);
-	      gtk_object_sink (GTK_OBJECT (info->proxy));
+	      g_object_ref_sink (info->proxy);
 	      gtk_widget_set_name (info->proxy, info->name);
 	      
 	      gtk_toolbar_insert (GTK_TOOLBAR (toolbar),
-				  GTK_TOOL_ITEM (info->proxy), pos);
-	    }
+	  		          GTK_TOOL_ITEM (info->proxy), pos);
+            }
 	}
       else
 	{
@@ -2533,14 +2613,11 @@ update_node (GtkUIManager *self,
 	  gtk_action_connect_proxy (action, info->proxy);
 	}
 
-      /* FIXME: we must trigger the notify::tooltip handler, since 
-       * tooltips on toolitems can't be set before the toolitem 
-       * is added to the toolbar.
-       */
-      g_object_notify (G_OBJECT (action), "tooltip");
-
-      g_signal_connect (info->proxy, "notify::visible",
-			G_CALLBACK (update_smart_separators), NULL);
+      if (info->proxy)
+        {
+          g_signal_connect (info->proxy, "notify::visible",
+			    G_CALLBACK (update_smart_separators), NULL);
+        }
       break;
     case NODE_TYPE_SEPARATOR:
       if (NODE_INFO (node->parent)->type == NODE_TYPE_TOOLBAR ||
@@ -2549,7 +2626,8 @@ update_node (GtkUIManager *self,
 	  GtkWidget *toolbar;
 	  gint pos;
 	  gint separator_mode;
-	  
+	  GtkToolItem *item;
+
 	  if (GTK_IS_SEPARATOR_TOOL_ITEM (info->proxy))
 	    {
 	      gtk_container_remove (GTK_CONTAINER (info->proxy->parent),
@@ -2559,28 +2637,26 @@ update_node (GtkUIManager *self,
 	    }
 	  
 	  if (find_toolbar_position (node, &toolbar, &pos))
-	    {
-	      GtkToolItem *item = gtk_separator_tool_item_new ();
+            {
+	      item  = gtk_separator_tool_item_new ();
 	      gtk_toolbar_insert (GTK_TOOLBAR (toolbar), item, pos);
 	      info->proxy = GTK_WIDGET (item);
-	      g_object_ref (info->proxy);
-	      gtk_object_sink (GTK_OBJECT (info->proxy));
+	      g_object_ref_sink (info->proxy);
 	      gtk_widget_set_no_show_all (info->proxy, TRUE);
 	      if (info->expand)
-		{
-		  gtk_tool_item_set_expand (GTK_TOOL_ITEM (item), TRUE);
-		  gtk_separator_tool_item_set_draw
-		    (GTK_SEPARATOR_TOOL_ITEM (item), FALSE);
-		  separator_mode = SEPARATOR_MODE_VISIBLE;
-		}
-		else
-		  separator_mode = SEPARATOR_MODE_SMART;
-
+	        {
+	          gtk_tool_item_set_expand (GTK_TOOL_ITEM (item), TRUE);
+	          gtk_separator_tool_item_set_draw (GTK_SEPARATOR_TOOL_ITEM (item), FALSE);
+	          separator_mode = SEPARATOR_MODE_VISIBLE;
+	        }
+	      else
+	        separator_mode = SEPARATOR_MODE_SMART;
+	  
 	      g_object_set_data (G_OBJECT (info->proxy),
-				 "gtk-separator-mode",
-				 GINT_TO_POINTER (separator_mode));
+			         I_("gtk-separator-mode"),
+			         GINT_TO_POINTER (separator_mode));
 	      gtk_widget_show (info->proxy);
-	    }
+            }
 	}
       else
 	{
@@ -2597,17 +2673,16 @@ update_node (GtkUIManager *self,
 	  
 	  if (find_menu_position (node, &menushell, &pos))
 	    {
-	      info->proxy = gtk_separator_menu_item_new ();
-	      g_object_ref (info->proxy);
-	      gtk_object_sink (GTK_OBJECT (info->proxy));
+              info->proxy = gtk_separator_menu_item_new ();
+	      g_object_ref_sink (info->proxy);
 	      gtk_widget_set_no_show_all (info->proxy, TRUE);
 	      g_object_set_data (G_OBJECT (info->proxy),
-				 "gtk-separator-mode",
-				 GINT_TO_POINTER (SEPARATOR_MODE_SMART));
+			         I_("gtk-separator-mode"),
+			         GINT_TO_POINTER (SEPARATOR_MODE_SMART));
 	      gtk_menu_shell_insert (GTK_MENU_SHELL (menushell),
 				     info->proxy, pos);
 	      gtk_widget_show (info->proxy);
-	    }
+            }
 	}
       break;
     case NODE_TYPE_ACCELERATOR:
@@ -2682,9 +2757,7 @@ do_updates (GtkUIManager *self)
 static gboolean
 do_updates_idle (GtkUIManager *self)
 {
-  GDK_THREADS_ENTER ();
   do_updates (self);
-  GDK_THREADS_LEAVE ();
 
   return FALSE;
 }
@@ -2695,7 +2768,8 @@ queue_update (GtkUIManager *self)
   if (self->private_data->update_tag != 0)
     return;
 
-  self->private_data->update_tag = g_idle_add ((GSourceFunc)do_updates_idle, 
+  self->private_data->update_tag = gdk_threads_add_idle (
+					       (GSourceFunc)do_updates_idle, 
 					       self);
 }
 
@@ -2761,35 +2835,43 @@ mark_node_dirty (GNode *node)
     NODE_INFO (p)->dirty = TRUE;  
 }
 
-static const gchar *const open_tag_format[] = {
-  "%*s<UNDECIDED",
-  "%*s<ui",
-  "%*s<menubar",
-  "%*s<menu",
-  "%*s<toolbar",
-  "%*s<placeholder",
-  "%*s<placeholder",
-  "%*s<popup",
-  "%*s<menuitem",
-  "%*s<toolitem",
-  "%*s<separator",
-  "%*s<accelerator"
-};
+static const gchar *
+open_tag_format (NodeType type)
+{
+  switch (type)
+    {
+    case NODE_TYPE_UNDECIDED: return "%*s<UNDECIDED"; 
+    case NODE_TYPE_ROOT: return "%*s<ui"; 
+    case NODE_TYPE_MENUBAR: return "%*s<menubar";
+    case NODE_TYPE_MENU: return "%*s<menu";
+    case NODE_TYPE_TOOLBAR: return "%*s<toolbar";
+    case NODE_TYPE_MENU_PLACEHOLDER:
+    case NODE_TYPE_TOOLBAR_PLACEHOLDER: return "%*s<placeholder";
+    case NODE_TYPE_POPUP: return "%*s<popup";
+    case NODE_TYPE_MENUITEM: return "%*s<menuitem";
+    case NODE_TYPE_TOOLITEM: return "%*s<toolitem";
+    case NODE_TYPE_SEPARATOR: return "%*s<separator";
+    case NODE_TYPE_ACCELERATOR: return "%*s<accelerator";
+    default: return NULL;
+    }
+}
 
-static const gchar *const close_tag_format[] = {
-  "%*s</UNDECIDED>\n",
-  "%*s</ui>\n",
-  "%*s</menubar>\n",
-  "%*s</menu>\n",
-  "%*s</toolbar>\n",
-  "%*s</placeholder>\n",
-  "%*s</placeholder>\n",
-  "%*s</popup>\n",
-  NULL,
-  NULL,
-  NULL,
-  NULL
-};
+static const gchar *
+close_tag_format (NodeType type)
+{
+  switch (type)
+    {
+    case NODE_TYPE_UNDECIDED: return "%*s</UNDECIDED>\n";
+    case NODE_TYPE_ROOT: return "%*s</ui>\n";
+    case NODE_TYPE_MENUBAR: return "%*s</menubar>\n";
+    case NODE_TYPE_MENU: return "%*s</menu>\n";
+    case NODE_TYPE_TOOLBAR: return "%*s</toolbar>\n";
+    case NODE_TYPE_MENU_PLACEHOLDER:
+    case NODE_TYPE_TOOLBAR_PLACEHOLDER: return "%*s</placeholder>\n";
+    case NODE_TYPE_POPUP: return "%*s</popup>\n";
+    default: return NULL;
+    }
+}
 
 static void
 print_node (GtkUIManager *self,
@@ -2799,11 +2881,15 @@ print_node (GtkUIManager *self,
 {
   Node  *mnode;
   GNode *child;
+  const gchar *open_fmt;
+  const gchar *close_fmt;
 
   mnode = node->data;
 
-  g_string_append_printf (buffer, open_tag_format[mnode->type],
-			  indent_level, "");
+  open_fmt = open_tag_format (mnode->type);
+  close_fmt = close_tag_format (mnode->type);
+
+  g_string_append_printf (buffer, open_fmt, indent_level, "");
 
   if (mnode->type != NODE_TYPE_ROOT)
     {
@@ -2815,15 +2901,56 @@ print_node (GtkUIManager *self,
 				g_quark_to_string (mnode->action_name));
     }
 
-  g_string_append (buffer,
-                   close_tag_format[mnode->type] ? ">\n" : "/>\n");
+  g_string_append (buffer, close_fmt ? ">\n" : "/>\n");
 
   for (child = node->children; child != NULL; child = child->next)
     print_node (self, child, indent_level + 2, buffer);
 
-  if (close_tag_format[mnode->type])
-    g_string_append_printf (buffer, close_tag_format[mnode->type],
-                            indent_level, "");
+  if (close_fmt)
+    g_string_append_printf (buffer, close_fmt, indent_level, "");
+}
+
+static gboolean
+gtk_ui_manager_buildable_custom_tag_start (GtkBuildable  *buildable,
+					   GtkBuilder    *builder,
+					   GObject       *child,
+					   const gchar   *tagname,
+					   GMarkupParser *parser,
+					   gpointer      *data)
+{
+  if (child)
+    return FALSE;
+
+  if (strcmp (tagname, "ui") == 0)
+    {
+      ParseContext *ctx;
+
+      ctx = g_new0 (ParseContext, 1);
+      ctx->state = STATE_START;
+      ctx->self = GTK_UI_MANAGER (buildable);
+      ctx->current = NULL;
+      ctx->merge_id = gtk_ui_manager_new_merge_id (GTK_UI_MANAGER (buildable));
+
+      *data = ctx;
+      *parser = ui_parser;
+
+      return TRUE;
+    }
+
+  return FALSE;
+
+}
+
+static void
+gtk_ui_manager_buildable_custom_tag_end (GtkBuildable *buildable,
+					 GtkBuilder   *builder,
+					 GObject      *child,
+					 const gchar  *tagname,
+					 gpointer     *data)
+{
+  queue_update (GTK_UI_MANAGER (buildable));
+  g_object_notify (G_OBJECT (buildable), "ui");
+  g_free (data);
 }
 
 /**
